# coding=utf8
"""Resize

Common functions for resizing dimensions to fit, to crop, etc
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-11"

def crop(w, h, bw, bh):
	"""Crop

	Makes sure one side fits and crops the other

	Arguments:
		w (int): The current width
		h (int): The current height
		bw (int): The boundary width
		bh (int): The boundary height

	Returns:
		dict
	"""

	# Init the return
	dRet = {}

	# Easier to work with floats
	w = float(w)
	h = float(h)

	# If the image is already smaller, make it bigger
	if w < bw or h < bh:

		# Which is the side that needs to grow more?
		if (bw / w) > (bh / h):
			dRet['w'] = bw
			dRet['h'] = int(round(bw * (h / w)))
		else:
			dRet['w'] = int(round(bh * (w / h)))
			dRet['h'] = bh

	# Else, make it smaller
	else:

		# Which is the side that needs to shrink less?
		if (w / bw) > (h / bh):
			dRet['w'] = int(round(bh * (w / h)))
			dRet['h'] = bh
		else:
			dRet['w'] = bw
			dRet['h'] = int(round(bw * (h / w)))

	# Return the new width and height
	return dRet

def fit(w, h, bw, bh):
	"""Fit

	Makes sure one side fits and makes the other smaller than necessary

	Arguments:
		w (int): The current width
		h (int): The current height
		bw (int): The boundary width
		bh (int): The boundary height

	Returns:
		list [w, h]
	"""

	# Init the return
	dRet = {}

	# Easier to work with floats
	w = float(w)
	h = float(h)

	# If the image is already smaller, make it bigger
	if w < bw and h < bh:

		# Figure out the larger side
		if (bw / w) > (bh / h):
			dRet['w'] = int(round(bh * (w / h)))
			dRet['h'] = bh
		else:
			dRet['w'] = bw
			dRet['h'] = int(round(bw * (h / w)))

	# Else, make it smaller
	else:

		# Figure out the larger side
		if (w / bw) > (h / bh):
			dRet['w'] = bw
			dRet['h'] = int(round(bw * (h / w)))
		else:
			dRet['w'] = int(round(bh * (w / h)))
			dRet['h'] = bh

	# Return the new width and height
	return dRet

def region(w, h, bw, bh):
	"""Region

	Returns a new set of region points based on a current width and height and
	the bounding box

	Arguments:
		w (int): The current width
		h (int): The current height
		bw (int): The boundary width
		bh (int): The boundary height

	Returns:
		dict
	"""

	# Return
	dRet = {}

	# If the current width is larger than the bounds width
	if w > bw:

		dRet['x'] = int(round((w - bw) / 2.0))
		dRet['y'] = 0
		dRet['w'] = int(bw + round((w - bw) / 2.0))
		dRet['h'] = bh

	# Else if the current height is larger than the bounds height
	else:

		dRet['x'] = 0
		dRet['y'] = int(round((h - bh) / 2.0))
		dRet['w'] = bw
		dRet['h'] = int(bh + round((h - bh) / 2.0))

	# Return the region
	return dRet



# coding=utf8
""" Conf Module

Used to set and get conf (json) files
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-11"

# Python imports
import sys

# Module imports
from . import DictHelper, JSON

# Confs available
_dmConfs = {}

def get(key = None, default = None, conf = '_'):
	"""Get

	Gets a single value from a conf, returns default if not found

	Arguments:
		key (string|tuple): A tuple to find a child key, or a string to find a
			parent key
		default (mixed): What to return if the key isn't found
		conf (string): The conf to fetch from

	Returns:
		mixed
	"""

	# Import the confs
	global _dmConfs

	# If the conf doesn't exist, return the default
	if conf not in _dmConfs:
		return default

	# If the key is none, return the entire conf
	if key is None:
		return _dmConfs[conf]

	# If the key is a string
	if isinstance(key, str):

		# If the key doesn't exist in the conf, return the default
		if key not in _dmConfs[conf]:
			return default

		# Return the key from the conf
		return _dmConfs[conf][key]

	# If the key is a tuple
	if isinstance(key, tuple):

		# Get the first level
		mRet = _dmConfs[conf]

		# Loop through and find the value
		for i in range(0, len(key)):

			# If we don't have a dict, return the default
			if not isinstance(mRet, dict):
				return default

			# If the key doesn't exist at this level, return the default
			if key[i] not in mRet:
				return default

			# Go deeper
			mRet = mRet[key[i]]

		# Return the value
		return mRet

	# Else, we got something bad
	raise ValueError('%s is not a valid value for key argument of %s' % (str(key), sys._getframe().f_code.co_name))

def load(file, conf = '_'):
	"""Load

	Loads a conf into the specified conf, or 'default', if not set

	Argurments:
		file (str): A file to be loaded into the conf
		conf (string): The conf to store in

	Returns:
		None
	"""

	# Import the confs
	global _dmConfs

	# Open the file and store the contents
	_dmConfs[conf] = JSON.load(file)

def load_merge(file, conf = '_'):
	"""Load Merge

	Loads a conf and merges it into whatever already exists

	Args:
		file (str): A file to be loaded
		conf (string): The conf to update

	Returns:
		None
	"""

	# Import the confs
	global _dmConfs

	# If we have no conf yet, just store it
	if conf not in _dmConfs:
		_dmConfs[conf] = JSON.load(file)

	# Else, combine the data
	else:
		_dmConfs[conf] = DictHelper.combine(_dmConfs[conf], JSON.load(file))

def set(key, val, conf = '_'):
	"""Set

	Sets a key in the given conf, returns the previous value

	Arguments:
		key (string|tuple): A tuple to set a child key, or a string to set a
			parent key
		val (mixed): The value to set for the key
		conf (str): The conf to set the key for

	Returns:
		mixed
	"""

	# Import the confs
	global _dmConfs

	# If the conf doesn't exist, create it
	if conf not in _dmConfs:
		_dmConfs[conf]

	# If the key is none, set the entire conf
	if key is None:
		_dmConfs[conf] = val
		return None

	# If the key is a string, set it
	if isinstance(key, str):
		mRet = (key in _dmConfs[conf] and _dmConfs[conf][key] or None)
		_dmConfs[conf][key] = val
		return mRet

	# If we got a tuple
	if isinstance(key, tuple):

		# Set the first level
		mSet = _dmConfs[conf]

		# Get the length of keys
		iLen = len(key)

		# Go through each level
		for i in range(0, iLen):

			# If we are on the last level
			if i+1 == iLen:
				mRet = mSet[key[i]]
				mSet[key[i]] = val
				return mRet

			# Else, we are still going deeper
			else:

				# If the level doesn't exist
				if key[i] not in mSet:
					mSet[key[i]] = {}

				# Set the level
				mSet = mSet[key[i]]

	# Else, we got something bad
	raise ValueError('%s is not a valid value for key argument of %s' % (str(key), sys._getframe().f_code.co_name))



# coding=utf8
""" Templates Module

Holds methods for handling templates
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-17"

# Pip imports
from jinja2 import Environment, FileSystemLoader
import pdfkit

# jinja2 environment
__moEnv = None

# preloaded templates
__mdTpls = {}

def init(folder):
	"""Init

	Initialise the Templates module

	Arguments:
		folder (str): The folder to find templates in

	Returns:
		None
	"""

	# Import the module variable
	global __moEnv

	# Init the template environment
	__moEnv = Environment(
		loader=FileSystemLoader(folder, followlinks=True, encoding='utf-8'),
		lstrip_blocks=True,
		trim_blocks=True
	)

def generate(tpl, data = {}, locale = 'en_US', pdf = False):
	"""Generate

	Generate content from a template and return it. Please note, when generating
	PDFs, wkhtmltopdf must be be installed on your system

	Arguments:
		tpl (str): The template to load
		data (dict): The keys/values to use in the template
		locale (str): The language to use, i.e. which folder to get the
			template from
		pdf (bool): Set to True to return a PDF of the generated template

	Returns:
		str
	"""

	# Import the module variables
	global __moEnv, __mdTpls

	# Generate the full path
	sPath = "%s/%s" % (locale, tpl)

	# If we don't already have the template
	if sPath not in __mdTpls:
		__mdTpls[sPath] = __moEnv.get_template(sPath)

	# Generate the template
	sContent = __mdTpls[sPath].render(**data)

	# If we want a PDF
	if pdf:
		return pdfkit.from_string(sContent, False)

	# Return the content
	return sContent



# coding=utf8
""" StrHelper Module

Several useful helper methods for use with strings
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-11"

# Python imports
from base64 import b64encode, b64decode
from random import randint
import sys

# The sets available for the random function
_mdRandomSets = {
	"0x":	"0123456789abcdef",
	"0":	"01234567",
	"10":	"0123456789",
	"10*":  "123456789",
	"az":	"abcdefghijklmnopqrstuvwxyz",
	"az*":	"abcdefghijkmnopqrstuvwxyz",
	"AZ":	"ABCDEFGHIJKLMNOPQRSTUVWXYZ",
	"AZ*":	"ABCDEFGHJKLMNPQRSTUVWXYZ",
	"aZ":	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
	"aZ*":	"abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ",
	"!":	"!@#$%^&*-_+.?",
	"!*":	"!@$^*-_."
}

def bytes_human(num, suffix='B'):
	"""Bytes Human

	Returns the size of bytes in the closest binary prefix so that they are
	clearly understood by humans

	Arguments:
		num (uint): The bytes to convert to human readable

	Returns:
		str
	"""

	for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:

		if abs(num) < 1024.0:
			return "%3.1f%sB" % (num, unit)

		num /= 1024.0

	return "%.1fYiB" % num

def decrypt(key, val):
	"""Decrypt

	Decrypts a string using the key and returns it. Key must be in multiples of
	16 bytes

	Arguments:
		key (str): A key that was used to encrypt the original value
		val (str): The value to decrypt

	Returns:
		str
	"""

	# Load PyCrypto
	try:
		from Cryptodome.Cipher import AES
	except Exception as e:
		print('%s\n' % str(e))
		return None

	# base64 decode the value, this can explode because of padding errors
	try: val = b64decode(val)
	except TypeError: return None

	# Recreate the IV
	sIV = val[:16]

	# Strip out the IV from the encrypted value
	val = val[16:]

	# Create the cipher and store the decrypted value
	oC = AES.new(key, AES.MODE_CFB, sIV, segment_size=128)

	# Return the decrypted value
	return oC.decrypt(val)

def digits(val):
	"""Digits

	Returns only the digits in the string

	Arguments:
		val (str): The string to strip all non-digit characters from

	Returns:
		str
	"""

	# Init list of valid characters
	lRet = []

	# Go through each character in the string and only keep it if it's a digit
	for c in val:
		if c.isdigit():
			lRet.append(c)

	# Return the new string
	return ''.join(lRet)

def encrypt(key, val):
	"""Encrypt

	Encrypts a string using the passed key and returns it. Key must be in
	multiples of 16 bytes

	Arguments:
		key (str): The key used to encrypt the value
		val (str): The value to encrypt and return

	Returns:
		str
	"""

	# Try to load PyCrypto
	try:
		from Crypto import Random
		from Cryptodome.Cipher import AES
	except Exception as e:
		print('%s\n' % str(e))
		return None

	# Generate an IV
	sIV = Random.new().read(AES.block_size)

	# Create a new cipher using the key and the IV
	oC = AES.new(key, AES.MODE_CFB, sIV, segment_size=128)

	# Encrypt the value
	val = oC.encrypt(val)

	# Add the IV
	val = '%s%s' % (sIV, val)

	# Return the entire thing as a base 64 encoded string
	return b64encode(val)

def normalize(val):
	"""Normalize

	Replaces all special alpha characters with their ascii equivalent

	Args:
		val (str): The text to normalize

	Returns:
		str
	"""
	return strtr(val, {
		'Ъ': "'", 'ъ': "'", 'Ь': "'", 'ь': "'",

		'Á': 'A', 'Ă': 'A', 'Ắ': 'A', 'Ặ': 'A', 'Ằ': 'A', 'Ẳ': 'A', 'Ẵ': 'A',
		'Ǎ': 'A', 'Â': 'A', 'Ấ': 'A', 'Ậ': 'A', 'Ầ': 'A', 'Ẩ': 'A', 'Ẫ': 'A',
		'Ä': 'A', 'Ǟ': 'A', 'Ȧ': 'A', 'Ǡ': 'A', 'Ạ': 'A', 'Ȁ': 'A', 'À': 'A',
		'Ả': 'A', 'Ȃ': 'A', 'Ā': 'A', 'Ą': 'A', 'Å': 'A', 'Ǻ': 'A', 'Ḁ': 'A',
		'Ⱥ': 'A', 'Ã': 'A', 'Ɐ': 'A', 'ᴀ': 'A',
		 'á': 'a', 'ă': 'a', 'ắ': 'a', 'ặ': 'a', 'ằ': 'a', 'ẳ': 'a', 'ẵ': 'a',
		 'ǎ': 'a', 'â': 'a', 'ấ': 'a', 'ậ': 'a', 'ầ': 'a', 'ẩ': 'a', 'ẫ': 'a',
		 'ä': 'a', 'ǟ': 'a', 'ȧ': 'a', 'ǡ': 'a', 'ạ': 'a', 'ȁ': 'a', 'à': 'a',
		 'ả': 'a', 'ȃ': 'a', 'ā': 'a', 'ą': 'a', 'ᶏ': 'a', 'ẚ': 'a', 'å': 'a',
		 'ǻ': 'a', 'ḁ': 'a', 'ⱥ': 'a', 'ã': 'a', 'ɐ': 'a', 'ₐ': 'a', 'А': 'a',
		 'а': 'a',

		'Ꜳ': 'AA', 'Æ': 'AE', 'Ǽ': 'AE', 'Ǣ': 'AE', 'ᴁ': 'AE', 'Ꜵ': 'AO',
		'Ꜷ': 'AU', 'Ꜹ': 'AV', 'Ꜻ': 'AV', 'Ꜽ': 'AY',
		'ꜳ': 'aa', 'æ': 'ae', 'ǽ': 'ae', 'ǣ': 'ae', 'ᴂ': 'ae', 'ꜵ': 'ao',
		'ꜷ': 'au', 'ꜹ': 'av', 'ꜻ': 'av', 'ꜽ': 'ay',

		'Ḃ': 'B', 'Ḅ': 'B', 'Ɓ': 'B', 'Ḇ': 'B', 'Ƀ': 'B', 'Ƃ': 'B', 'ʙ': 'B',
		'ᴃ': 'B', 'Б': 'B',
		'ḃ': 'b', 'ḅ': 'b', 'ɓ': 'b', 'ḇ': 'b', 'ᵬ': 'b', 'ᶀ': 'b', 'ƀ': 'b',
		'ƃ': 'b', 'б': 'b',

		'Ć': 'C', 'Č': 'C', 'Ç': 'C', 'Ḉ': 'C', 'Ĉ': 'C', 'Ċ': 'C', 'Ƈ': 'C',
		'Ȼ': 'C', 'Ꜿ': 'C', 'ᴄ': 'C',
		'ć': 'c', 'č': 'c', 'ç': 'c', 'ḉ': 'c', 'ĉ': 'c', 'ɕ': 'c', 'ċ': 'c',
		'ƈ': 'c', 'ȼ': 'c', 'ↄ': 'c', 'ꜿ': 'c',

		'Ч': 'CH',
		'ч': 'ch',

		'Ď': 'D', 'Ḑ': 'D', 'Ḓ': 'D', 'Ḋ': 'D', 'Ḍ': 'D', 'Ɗ': 'D', 'Ḏ': 'D',
		'ǲ': 'D', 'ǅ': 'D', 'Đ': 'D', 'Ƌ': 'D', 'Ꝺ': 'D', 'ᴅ': 'D', 'Д': 'D',
		'ď': 'd', 'ḑ': 'd', 'ḓ': 'd', 'ȡ': 'd', 'ḋ': 'd', 'ḍ': 'd', 'ɗ': 'd',
		'ᶑ': 'd', 'ḏ': 'd', 'ᵭ': 'd', 'ᶁ': 'd', 'đ': 'd', 'ɖ': 'd', 'ƌ': 'd',
		'ꝺ': 'd', 'д': 'd',

		'Ǳ': 'DZ', 'Ǆ': 'DZ',
		'ǳ': 'dz', 'ǆ': 'dz',

		'É': 'E', 'Ĕ': 'E', 'Ě': 'E', 'Ȩ': 'E', 'Ḝ': 'E', 'Ê': 'E', 'Ế': 'E',
		'Ệ': 'E', 'Ề': 'E', 'Ể': 'E', 'Ễ': 'E', 'Ḙ': 'E', 'Ë': 'E', 'Ė': 'E',
		'Ẹ': 'E', 'Ȅ': 'E', 'È': 'E', 'Ẻ': 'E', 'Ȇ': 'E', 'Ē': 'E', 'Ḗ': 'E',
		'Ḕ': 'E', 'Ę': 'E', 'Ɇ': 'E', 'Ẽ': 'E', 'Ḛ': 'E', 'Ɛ': 'E', 'Ǝ': 'E',
		'ᴇ': 'E', 'ⱻ': 'E', 'Е': 'E', 'Э': 'E',
		'é': 'e', 'ĕ': 'e', 'ě': 'e', 'ȩ': 'e', 'ḝ': 'e', 'ê': 'e', 'ế': 'e',
		'ệ': 'e', 'ề': 'e', 'ể': 'e', 'ễ': 'e', 'ḙ': 'e', 'ë': 'e', 'ė': 'e',
		'ẹ': 'e', 'ȅ': 'e', 'è': 'e', 'ẻ': 'e', 'ȇ': 'e', 'ē': 'e', 'ḗ': 'e',
		'ḕ': 'e', 'ⱸ': 'e', 'ę': 'e', 'ᶒ': 'e', 'ɇ': 'e', 'ẽ': 'e', 'ḛ': 'e',
		'ɛ': 'e', 'ᶓ': 'e', 'ɘ': 'e', 'ǝ': 'e', 'ₑ': 'e', 'е': 'e', 'э': 'e',

		'Ꝫ': 'ET',
		'ꝫ': 'et',

		'Ḟ': 'F', 'Ƒ': 'F', 'Ꝼ': 'F', 'ꜰ': 'F', 'Ф': 'F',
		'ḟ': 'f', 'ƒ': 'f', 'ᵮ': 'f', 'ᶂ': 'f', 'ꝼ': 'f', 'ф': 'f',

		'ﬀ': 'ff', 'ﬃ': 'ffi', 'ﬄ': 'ffl', 'ﬁ': 'fi', 'ﬂ': 'fl',

		'Ǵ': 'G', 'Ğ': 'G', 'Ǧ': 'G', 'Ģ': 'G', 'Ĝ': 'G', 'Ġ': 'G', 'Ɠ': 'G',
		'Ḡ': 'G', 'Ǥ': 'G', 'Ᵹ': 'G', 'ɢ': 'G', 'ʛ': 'G', 'Г': 'G',
		'ǵ': 'g', 'ğ': 'g', 'ǧ': 'g', 'ģ': 'g', 'ĝ': 'g', 'ġ': 'g', 'ɠ': 'g',
		'ḡ': 'g', 'ᶃ': 'g', 'ǥ': 'g', 'ᵹ': 'g', 'ɡ': 'g', 'ᵷ': 'g', 'г': 'g',

		'Ḫ': 'H', 'Ȟ': 'H', 'Ḩ': 'H', 'Ĥ': 'H', 'Ⱨ': 'H', 'Ḧ': 'H', 'Ḣ': 'H',
		'Ḥ': 'H', 'Ħ': 'H', 'ʜ': 'H', 'Х': 'H',
		'ḫ': 'h', 'ȟ': 'h', 'ḩ': 'h', 'ĥ': 'h', 'ⱨ': 'h', 'ḧ': 'h', 'ḣ': 'h',
		'ḥ': 'h', 'ɦ': 'h', 'ẖ': 'h', 'ħ': 'h', 'ɥ': 'h', 'ʮ': 'h', 'ʯ': 'h',
		'х': 'h',

		'ƕ': 'hv',

		'Í': 'I', 'Ĭ': 'I', 'Ǐ': 'I', 'Î': 'I', 'Ï': 'I', 'Ḯ': 'I', 'İ': 'I',
		'Ị': 'I', 'Ȉ': 'I', 'Ì': 'I', 'Ỉ': 'I', 'Ȋ': 'I', 'Ī': 'I', 'Į': 'I',
		'Ɨ': 'I', 'Ĩ': 'I', 'Ḭ': 'I', 'ɪ': 'I', 'Й': 'I', 'Ы': 'I', 'И': 'I',
		'ı': 'i', 'í': 'i', 'ĭ': 'i', 'ǐ': 'i', 'î': 'i', 'ï': 'i', 'ḯ': 'i',
		'ị': 'i', 'ȉ': 'i', 'ì': 'i', 'ỉ': 'i', 'ȋ': 'i', 'ī': 'i', 'į': 'i',
		'ᶖ': 'i', 'ɨ': 'i', 'ĩ': 'i', 'ḭ': 'i', 'ᴉ': 'i', 'ᵢ': 'i', 'й': 'i',
		'ы': 'i', 'и': 'i',

		'Ĳ': 'IJ', 'Ꝭ': 'IS',
		'ĳ': 'ij', 'ꝭ': 'is',

		'Ĵ': 'J', 'Ɉ': 'J', 'ᴊ': 'J',
		'ȷ': 'j', 'ɟ': 'j', 'ʄ': 'j', 'ǰ': 'j', 'ĵ': 'j', 'ʝ': 'j', 'ɉ': 'j',
		'ⱼ': 'j',

		'Ḱ': 'K', 'Ǩ': 'K', 'Ķ': 'K', 'Ⱪ': 'K', 'Ꝃ': 'K', 'Ḳ': 'K', 'Ƙ': 'K',
		'Ḵ': 'K', 'Ꝁ': 'K', 'Ꝅ': 'K', 'ᴋ': 'K', 'К': 'K',
		'ḱ': 'k', 'ǩ': 'k', 'ķ': 'k', 'ⱪ': 'k', 'ꝃ': 'k', 'ḳ': 'k', 'ƙ': 'k',
		'ḵ': 'k', 'ᶄ': 'k', 'ꝁ': 'k', 'ꝅ': 'k', 'ʞ': 'k', 'к': 'k',

		'Ĺ': 'L', 'Ƚ': 'L', 'Ľ': 'L', 'Ļ': 'L', 'Ḽ': 'L', 'Ḷ': 'L', 'Ḹ': 'L',
		'Ⱡ': 'L', 'Ꝉ': 'L', 'Ḻ': 'L', 'Ŀ': 'L', 'Ɫ': 'L', 'ǈ': 'L', 'Ł': 'L',
		'Ꞁ': 'L', 'ʟ': 'L', 'ᴌ': 'L', 'Л': 'L',
		'ĺ': 'l', 'ƚ': 'l', 'ɬ': 'l', 'ľ': 'l', 'ļ': 'l', 'ḽ': 'l', 'ȴ': 'l',
		'ḷ': 'l', 'ḹ': 'l', 'ⱡ': 'l', 'ꝉ': 'l', 'ḻ': 'l', 'ŀ': 'l', 'ɫ': 'l',
		'ᶅ': 'l', 'ɭ': 'l', 'ł': 'l', 'ꞁ': 'l', 'л': 'l',

		'Ǉ': 'LJ',
		'ǉ': 'lj',

		'Ḿ': 'M', 'Ṁ': 'M', 'Ṃ': 'M', 'Ɱ': 'M', 'Ɯ': 'M', 'ᴍ': 'M', 'М': 'M',
		'ḿ': 'm', 'ṁ': 'm', 'ṃ': 'm', 'ɱ': 'm', 'ᵯ': 'm', 'ᶆ': 'm', 'ɯ': 'm',
		'ɰ': 'm', 'м': 'm',

		'Ń': 'N', 'Ň': 'N', 'Ņ': 'N', 'Ṋ': 'N', 'Ṅ': 'N', 'Ṇ': 'N', 'Ǹ': 'N',
		'Ɲ': 'N', 'Ṉ': 'N', 'Ƞ': 'N', 'ǋ': 'N', 'Ñ': 'N', 'ɴ': 'N', 'ᴎ': 'N',
		'Н': 'N',
		'ń': 'n', 'ň': 'n', 'ņ': 'n', 'ṋ': 'n', 'ȵ': 'n', 'ṅ': 'n', 'ṇ': 'n',
		'ǹ': 'n', 'ɲ': 'n', 'ṉ': 'n', 'ƞ': 'n', 'ᵰ': 'n', 'ᶇ': 'n', 'ɳ': 'n',
		'ñ': 'n', 'н': 'n',

		'Ǌ': 'NJ',
		'ǌ': 'nj',

		'Ó': 'O', 'Ŏ': 'O', 'Ǒ': 'O', 'Ô': 'O', 'Ố': 'O', 'Ộ': 'O', 'Ồ': 'O',
		'Ổ': 'O', 'Ỗ': 'O', 'Ö': 'O', 'Ȫ': 'O', 'Ȯ': 'O', 'Ȱ': 'O', 'Ọ': 'O',
		'Ő': 'O', 'Ȍ': 'O', 'Ò': 'O', 'Ỏ': 'O', 'Ơ': 'O', 'Ớ': 'O', 'Ợ': 'O',
		'Ờ': 'O', 'Ở': 'O', 'Ỡ': 'O', 'Ȏ': 'O', 'Ꝋ': 'O', 'Ꝍ': 'O', 'Ō': 'O',
		'Ṓ': 'O', 'Ṑ': 'O', 'Ɵ': 'O', 'Ǫ': 'O', 'Ǭ': 'O', 'Ø': 'O', 'Ǿ': 'O',
		'Õ': 'O', 'Ṍ': 'O', 'Ṏ': 'O', 'Ȭ': 'O', 'Ɔ': 'O', 'ᴏ': 'O', 'ᴐ': 'O',
		'О': 'O',
		'ɵ': 'o', 'ó': 'o', 'ŏ': 'o', 'ǒ': 'o', 'ô': 'o', 'ố': 'o', 'ộ': 'o',
		'ồ': 'o', 'ổ': 'o', 'ỗ': 'o', 'ö': 'o', 'ȫ': 'o', 'ȯ': 'o', 'ȱ': 'o',
		'ọ': 'o', 'ő': 'o', 'ȍ': 'o', 'ò': 'o', 'ỏ': 'o', 'ơ': 'o', 'ớ': 'o',
		'ợ': 'o', 'ờ': 'o', 'ở': 'o', 'ỡ': 'o', 'ȏ': 'o', 'ꝋ': 'o', 'ꝍ': 'o',
		'ⱺ': 'o', 'ō': 'o', 'ṓ': 'o', 'ṑ': 'o', 'ǫ': 'o', 'ǭ': 'o', 'ø': 'o',
		'ǿ': 'o', 'õ': 'o', 'ṍ': 'o', 'ṏ': 'o', 'ȭ': 'o', 'ɔ': 'o', 'ᶗ': 'o',
		'ᴑ': 'o', 'ᴓ': 'o', 'ₒ': 'o', 'о': 'o',

		'Œ': 'OE', 'ɶ': 'OE', 'Ƣ': 'OI', 'Ꝏ': 'OO', 'Ȣ': 'OU', 'ᴕ': 'OU',
		'ᴔ': 'oe', 'œ': 'oe', 'ƣ': 'oi', 'ꝏ': 'oo', 'ȣ': 'ou',

		'Ṕ': 'P', 'Ṗ': 'P', 'Ꝓ': 'P', 'Ƥ': 'P', 'Ꝕ': 'P', 'Ᵽ': 'P', 'Ꝑ': 'P',
		'ᴘ': 'P', 'П': 'P',
		'ṕ': 'p', 'ṗ': 'p', 'ꝓ': 'p', 'ƥ': 'p', 'ᵱ': 'p', 'ᶈ': 'p', 'ꝕ': 'p',
		'ᵽ': 'p', 'ꝑ': 'p', 'п': 'p',

		'Ꝙ': 'Q', 'Ꝗ': 'Q',
		'ꝙ': 'q', 'ʠ': 'q', 'ɋ': 'q', 'ꝗ': 'q',

		'Ꞃ': 'R', 'Ŕ': 'R', 'Ř': 'R', 'Ŗ': 'R', 'Ṙ': 'R', 'Ṛ': 'R', 'Ṝ': 'R',
		'Ȑ': 'R', 'Ȓ': 'R', 'Ṟ': 'R', 'Ɍ': 'R', 'Ɽ': 'R', 'ʁ': 'R', 'ʀ': 'R',
		'ᴙ': 'R', 'ᴚ': 'R', 'Р': 'R',
		'ꞃ': 'r', 'ŕ': 'r', 'ř': 'r', 'ŗ': 'r', 'ṙ': 'r', 'ṛ': 'r', 'ṝ': 'r',
		'ȑ': 'r', 'ɾ': 'r', 'ᵳ': 'r', 'ȓ': 'r', 'ṟ': 'r', 'ɼ': 'r', 'ᵲ': 'r',
		'ᶉ': 'r', 'ɍ': 'r', 'ɽ': 'r', 'ɿ': 'r', 'ɹ': 'r', 'ɻ': 'r', 'ɺ': 'r',
		'ⱹ': 'r', 'ᵣ': 'r', 'р': 'r',

		'Ꞅ': 'S', 'Ś': 'S', 'Ṥ': 'S', 'Š': 'S', 'Ṧ': 'S', 'Ş': 'S', 'Ŝ': 'S',
		'Ș': 'S', 'Ṡ': 'S', 'Ṣ': 'S', 'Ṩ': 'S', 'ꜱ': 'S', 'С': 'S',
		'ꞅ': 's', 'ſ': 's', 'ẜ': 's', 'ẛ': 's', 'ẝ': 's', 'ś': 's', 'ṥ': 's',
		'š': 's', 'ṧ': 's', 'ş': 's', 'ŝ': 's', 'ș': 's', 'ṡ': 's', 'ṣ': 's',
		'ṩ': 's', 'ʂ': 's', 'ᵴ': 's', 'ᶊ': 's', 'ȿ': 's', 'с': 's',

		'Щ': 'SCH', 'Ш': 'SH',
		'щ': 'sch', 'ш': 'sh', 'ß': 'ss', 'ﬆ': 'st',

		'Ꞇ': 'T', 'Ť': 'T', 'Ţ': 'T', 'Ṱ': 'T', 'Ț': 'T', 'Ⱦ': 'T', 'Ṫ': 'T',
		'Ṭ': 'T', 'Ƭ': 'T', 'Ṯ': 'T', 'Ʈ': 'T', 'Ŧ': 'T', 'ᴛ': 'T', 'Т': 'T',
		'ꞇ': 't', 'ť': 't', 'ţ': 't', 'ṱ': 't', 'ț': 't', 'ȶ': 't', 'ẗ': 't',
		'ⱦ': 't', 'ṫ': 't', 'ṭ': 't', 'ƭ': 't', 'ṯ': 't', 'ᵵ': 't', 'ƫ': 't',
		'ʈ': 't', 'ŧ': 't', 'ʇ': 't', 'т': 't',

		'Ц': 'TS', 'Ꜩ': 'TZ',
		'ᵺ': 'th', 'ц': 'ts', 'ꜩ': 'tz',

		'Ú': 'U', 'Ŭ': 'U', 'Ǔ': 'U', 'Û': 'U', 'Ṷ': 'U', 'Ü': 'U', 'Ǘ': 'U',
		'Ǚ': 'U', 'Ǜ': 'U', 'Ǖ': 'U', 'Ṳ': 'U', 'Ụ': 'U', 'Ű': 'U', 'Ȕ': 'U',
		'Ù': 'U', 'Ủ': 'U', 'Ư': 'U', 'Ứ': 'U', 'Ự': 'U', 'Ừ': 'U', 'Ử': 'U',
		'Ữ': 'U', 'Ȗ': 'U', 'Ū': 'U', 'Ṻ': 'U', 'Ų': 'U', 'Ů': 'U', 'Ũ': 'U',
		'Ṹ': 'U', 'Ṵ': 'U', 'ᴜ': 'U', 'У': 'U',
		'ᴝ': 'u', 'ú': 'u', 'ŭ': 'u', 'ǔ': 'u', 'û': 'u', 'ṷ': 'u', 'ü': 'u',
		'ǘ': 'u', 'ǚ': 'u', 'ǜ': 'u', 'ǖ': 'u', 'ṳ': 'u', 'ụ': 'u', 'ű': 'u',
		'ȕ': 'u', 'ù': 'u', 'ủ': 'u', 'ư': 'u', 'ứ': 'u', 'ự': 'u', 'ừ': 'u',
		'ử': 'u', 'ữ': 'u', 'ȗ': 'u', 'ū': 'u', 'ṻ': 'u', 'ų': 'u', 'ᶙ': 'u',
		'ů': 'u', 'ũ': 'u', 'ṹ': 'u', 'ṵ': 'u', 'ᵤ': 'u', 'у': 'u',

		'ᵫ': 'ue', 'ꝸ': 'um',

		'Ʌ': 'V', 'Ꝟ': 'V', 'Ṿ': 'V', 'Ʋ': 'V', 'Ṽ': 'V', 'ᴠ': 'V', 'В': 'V',
		'ʌ': 'v', 'ⱴ': 'v', 'ꝟ': 'v', 'ṿ': 'v', 'ʋ': 'v', 'ᶌ': 'v', 'ⱱ': 'v',
		'ṽ': 'v', 'ᵥ': 'v', 'в': 'v',

		'Ꝡ': 'VY',
		'ꝡ': 'vy',

		'Ẃ': 'W', 'Ŵ': 'W', 'Ẅ': 'W', 'Ẇ': 'W', 'Ẉ': 'W', 'Ẁ': 'W', 'Ⱳ': 'W',
		'ᴡ': 'W',
		'ʍ': 'w', 'ẃ': 'w', 'ŵ': 'w', 'ẅ': 'w', 'ẇ': 'w', 'ẉ': 'w', 'ẁ': 'w',
		'ⱳ': 'w', 'ẘ': 'w',

		'Ẍ': 'X', 'Ẋ': 'X',
		'ẍ': 'x', 'ẋ': 'x', 'ᶍ': 'x', 'ₓ': 'x',

		'Ý': 'Y', 'Ŷ': 'Y', 'Ÿ': 'Y', 'Ẏ': 'Y', 'Ỵ': 'Y', 'Ỳ': 'Y', 'Ƴ': 'Y',
		'Ỷ': 'Y', 'Ỿ': 'Y', 'Ȳ': 'Y', 'Ɏ': 'Y', 'Ỹ': 'Y', 'ʏ': 'Y',
		'ʎ': 'y', 'ý': 'y', 'ŷ': 'y', 'ÿ': 'y', 'ẏ': 'y', 'ỵ': 'y', 'ỳ': 'y',
		'ƴ': 'y', 'ỷ': 'y', 'ỿ': 'y', 'ȳ': 'y', 'ẙ': 'y', 'ɏ': 'y', 'ỹ': 'y',

		'Ё': 'YO', 'Ю': 'YU', 'Я': 'Ya',
		'я': 'ya', 'ё': 'yo', 'ю': 'yu',

		'Ź': 'Z', 'Ž': 'Z', 'Ẑ': 'Z', 'Ⱬ': 'Z', 'Ż': 'Z', 'Ẓ': 'Z', 'Ȥ': 'Z',
		'Ẕ': 'Z', 'Ƶ': 'Z', 'ᴢ': 'Z', 'З': 'Z',
		'ź': 'z', 'ž': 'z', 'ẑ': 'z', 'ʑ': 'z', 'ⱬ': 'z', 'ż': 'z', 'ẓ': 'z',
		'ȥ': 'z', 'ẕ': 'z', 'ᵶ': 'z', 'ᶎ': 'z', 'ʐ': 'z', 'ƶ': 'z', 'ɀ': 'z',
		'з': 'z',

		'Ж': 'ZH',
		'ж': 'zh'
	})

def random(length = 8, sets='_aZ', duplicates=True):
	"""Random

	Generates a random string. By default this function will generate an 8
	character string using lowercase letters with possible repeating characters

	Arguments:
		length (int): Requested length of the string
		sets (str|str[]): A list of names from the standard sets, a string
			starting with an underscore representing one named set, or any other
			string to be used as an array of characters to chose from. If you
			want certain characters to have a greater chance of appearing, use
			them more times, e.g. twice the 'A's, "AABC", or three times the
			'B's, "ABBBC". Make sure not to turn off duplicates for this to be
			effective
		duplicates (bool): Defaults to True, allowing characters to be used
			more than once

	Sets:
		0x:		0123456789abcdef
		0:		01234567
		10:		0123456789
		az:		abcdefghijklmnopqrstuvwxyz
		az*:	abcdefghijkmnopqrstuvwxyz
		AZ:		ABCDEFGHIJKLMNOPQRSTUVWXYZ
		AZ*:	ABCDEFGHJKLMNPQRSTUVWXYZ
		aZ:		abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ
		aZ*:	abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ
		!:		!@#$%^&*-_+.?
		!*:		!@$%^*-_.

	Examples:
		> random(8, '_0x')
		"baadbeef"

	Returns:
		str
	"""

	# If the sets are a list
	if isinstance(sets, list):

		# If there is no count
		if not sets:
			raise ValueError('sets must contain at least one set name')

		# Init the string to be used as the allowed characters
		sChars = ''

		# Go through the list
		for s in sets:

			# If the set doesn't exist
			if s not in _mdRandomSets:
				raise ValueError('%s is not a valid set' % s)

			# Else, add it to the allowed characters
			sChars += _mdRandomSets[s]

	# Else if we have a string
	elif isinstance(sets, str):

		# If it starts with an underscore
		if sets[0] == '_':

			# If the set doesn't exist
			if sets[1:] not in _mdRandomSets:
				raise ValueError('%s is not a valid set for %s' % (sets[1:], sys._getframe().f_code.co_name))

			# Else, set it to the allowed characters
			sChars = _mdRandomSets[sets[1:]]

		# Else, use the string as is
		else:
			sChars = sets

	else:
		raise ValueError('%s is not a valid value for sets argument of %s' % (str(sets), sys._getframe().f_code.co_name))

	# Init the return variable
	sText = ''

	# Count the number of characters we can use
	iCount = len(sChars)

	# Create a [length] of random character
	i = 0
	while i < length:
		sFound = sChars[randint(0, iCount - 1)]
		bDup = sText.find(sFound)

		if duplicates or bDup == -1:
			sText += sFound
			i += 1

	# Return the generated string
	return sText

def strtr(text, table):
	"""String Translate

	Port of PHP strtr (string translate)

	Args:
		text (str): The string to translate
		table (dict): The translation table

	Returns:
		str
	"""
	text = str(text)
	buff = []
	i = 0
	n = len(text)
	while i < n:
		for s, r in table.items():
			if text[i:len(s)+i] == s:
				buff.append(r)
				i += len(s)
				break
		else:
			buff.append(text[i])
			i += 1

	return ''.join(buff)

def to_bool(t):
	"""To Bool

	Converts a string to a boolean value

	Arguments:
		t (str): The text to attempt to convert

	Raises
		ValueError

	Returns:
		bool
	"""

	# If we didn't get a string
	if not isinstance(t, str):
		return ValueError('t is not a string: %s', str(type(t)))

	# First, convert the string to lowercase
	t = t.lower()

	# If it's any true type value
	if t in ['1', 'on', 't', 'true', 'y', 'yes']:
		return True

	# Else, if it's any false type value
	elif t in ['0', 'f', 'false', 'n', 'no', 'off']:
		return False

	# Raise an exception
	raise ValueError('t is not a boolean representation: "%s"' % t)

def uuid_add_dashes(uuid):
	"""UUID Add Dashes

	Adds dashes back to a UUID that had them removed

	Arguments:
		uuid (str): The UUID to transform

	Returns:
		str
	"""
	return '%s-%s-%s-%s-%s' % (
		uuid[0:8],
		uuid[8:12],
		uuid[12:16],
		uuid[16:20],
		uuid[20:32]
	)

def uuid_strip_dashes(uuid):
	"""UUID Strip Dashes

	Removes the dashes from a UUID

	Arguments:
		uuid (str): The UUID to transform

	Returns:
		str
	"""
	return '%s%s%s%s%s' % (
		uuid[0:8],
		uuid[9:13],
		uuid[14:18],
		uuid[19:23],
		uuid[24:36]
	)



# coding=utf8
""" Dictionary Helper Module

Several useful helper methods for use with dicts
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-11"

# Python imports
import sys

def clone(src):
	"""Clone

	Goes through the dict and any child dicts copying the values so that we
	don't have any references

	Arguments:
		src (dict): The source dict

	Returns:
		dict
	"""

	# Check the argument
	if not isinstance(src, dict):
		raise ValueError('%s is not a valid value for src argument of %s' % (str(src), sys._getframe().f_code.co_name))

	# Initialise the new dict
	dRet = {}

	# Get each key of the source dict
	for k in src:

		# If the key points to another dict
		if isinstance(src[k], dict):

			# Call clone on it
			dRet[k] = clone(src[k])

		# Else if the key points to a list
		elif isinstance(src[k], list):

			# Use list magic to copy it
			dRet[k] = src[k][:]

		# Else it's a standard variable
		else:
			dRet[k] = src[k]

	# Return the new dict
	return dRet

def combine(first, second):
	"""Combine

	Generates a new dict by combining the two passed, values in second will
	overwrite values in first

	Arguments:
		first (dict): The dict to be changed/overwritten
		second (dict): The dict that will do the overwriting

	Returns:
		dict
	"""

	# Make sure both arguments are actual dicts
	if not isinstance(first, dict):
		raise ValueError('%s is not a valid value for first of %s' % (str(first), sys._getframe().f_code.co_name))
	if not isinstance(second, dict):
		raise ValueError('%s is not a valid value for second of %s' % (str(second), sys._getframe().f_code.co_name))

	# Copy the first dict
	dRet = clone(first)

	# Call merge to avoid duplicate code and return the cloned dict
	return merge(dRet, second)

# Evaluate function
def eval(src, contains):
	"""Eval(uate)

	Goes through a dict looking for keys from contains

	Arguments:
		src (dict): The dict we are evaluating
		contains (list): A list of values to check for, if the value is a dict
			rather than a string, epects keys to be keys pointing to further
			lists of keys

	Return:
		A list of errors, or None
	"""

	# Initialise the list of errors
	lErrs = []

	# Go through each contains value
	for s in contains:

		# If the value is a string
		if isinstance(s, str):

			# If value does not exist in the source
			if s not in src or (isinstance(src[s], str) and not src[s]):
				lErrs.append(s)

		# Else, if we got a dict
		elif isinstance(s, dict):

			# Go through the key/value pairs in the dict
			for k,v in s.items():

				# If the key doesn't exist in the source or has no value
				if k not in src or not src[k]:
					lErrs.append(k)

				# Else, check the children
				else:

					# Call the eval on the child dict
					lChildErrs = eval(src[k], v)

					# Add errors to the list
					if lChildErrs:
						for sErr in lChildErrs:
							lErrs.append(k + '.' + sErr)

		# We got an unknown type of key
		else:
			lErrs.append(str(s))

	# If there's any errors
	if lErrs:
		raise ValueError(*lErrs)

def keys_to_ints(src):
	"""Keys To Ints

	Recursively goes through a dictionary and converts all keys that are
	numeric but stored as strings to integers. Returns a new dict and doesn't
	alter the original.

	PLEASE NOTE: this method is not useful for classes, or anything complex, it
	is meant primarily for converting JSON objects which don't allow ints as
	keys. Passing a set, tuple, or iterable class will not result in the
	expected result

	Arguments:
		src (dict|list): The dict we are modifying, accepts lists in order to
							handle recursive following the data

	Returns:
		dict|list
	"""

	# If we got a dict
	if isinstance(src, dict):

		# Init the return value to an empty dict
		mRet = {}

		# Go through the each key of the source
		for k in src:

			# Is the value numeric?
			try: mK = int(k)
			except ValueError: mK = k

			# If we got a dict or list, recurse it
			if isinstance(src[k], (dict,list)):
				mRet[mK] = keys_to_ints(src[k])

			# Else, store as is
			else:
				mRet[mK] = src[k]

	# Else, if we got a list
	elif isinstance(src, list):

		# Init the result value to a list
		mRet = []

		# Go through each item in the list
		for i in range(len(src)):

			# If we got a dict or list, recurse it
			if isinstance(src[i], (dict,list)):
				mRet.append(keys_to_ints(src[i]))

			# Else, store as is
			else:
				mRet.append(src[k])

	# Else, raise an error
	else:
		raise ValueError('src of %s must be a dict or list, received %s' % (sys._getframe().f_code.co_name, str(type(src))))

	# Return the new data
	return mRet

def merge(first, second):
	"""Merge

	Overwrites the first dict by adding the values from the second. Returns the
	first for chaining / ease of use

	Arguments:
		first (dict): The dict to be changed/overwritten
		second (dict): The dict that will do the overwriting

	Returns:
		dict
	"""

	# Make sure both arguments are actual dicts
	if not isinstance(first, dict):
		raise ValueError('%s is not a valid value for first of %s' % (str(first), sys._getframe().f_code.co_name))
	if not isinstance(second, dict):
		raise ValueError('%s is not a valid value for second of %s' % (str(second), sys._getframe().f_code.co_name))

	# Get each key of the second dict
	for m in second:

		# If the value is another dict and it exists in first as well
		if isinstance(second[m], dict) and m in first and isinstance(first[m], dict):

			# Call merge
			merge(first[m], second[m])

		# else we overwrite the value as is
		else:
			first[m] = second[m]

	# Return the new dict
	return first


# coding=utf8
"""Record ReDB Module

Extends Record module to add support for RethinkDB tables
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-11"

# Python imports
from hashlib import md5
import sys
from time import sleep, time

# Pip imports
from rethinkdb import errors as rerrors, net as rnet, RethinkDB
r = RethinkDB()

# Module imports
from . import DictHelper, Record_Base

# List of available hosts
__mdHosts = {}

# defines
MAX_RETRIES = 3

# Backwards compatibility (and ease of use)
DuplicateException = Record_Base.DuplicateException
RecordException = Record_Base.RecordException

def _connect(host, error_count=0):
	"""Connect

	Internal module function to fetch a connection to a specific RethinkDB host

	Arguments:
		host (str): The name the host is stored using add_host()
		error_count (int): The number of times the function has failed

	Raises:
		ConnectionError

	Returns:
		rethinkdb.net.DefaultConnection
	"""

	# If no such host has been added
	if host not in __mdHosts:
		raise ValueError('no such host "%s"' % str(host))

	# Try to connect
	try:
		oCon = r.connect(**__mdHosts[host])

	# Check for driver errors
	except rerrors.RqlDriverError as e:

		# If there was an error, increment the error count
		error_count += 1

		# Raise an exception if we've tried enough times
		if error_count == MAX_RETRIES:
			raise ConnectionError(*e.args)

		# Sleep and try again
		else:
			sleep(1)
			return _connect(host, error_count)

	# Return the connection
	return oCon

class _with(object):
	"""_with

	Used with the special Python with method to create a connection that will
	always be closed regardless of exceptions
	"""

	def __init__(self, host):
		self.con = _connect(host)

	def __enter__(self):
		return self.con

	def __exit__(self, exc_type, exc_value, traceback):
		self.con.close()
		if exc_type is not None:
			return False

def add_host(name, info, update=False):
	"""Add Host

	Add a host that can be used by Records

	Arguments:
		name (str): The name that will be used to fetch the host credentials
		info (dict): The necessary credentials to connect to the host

	Returns:
		bool
	"""

	# If the info isn't already stored, or we want to overwrite it, store it
	if name not in __mdHosts or update:
		__mdHosts[name] = info
		return True

	# Nothing to do, not OK
	return False

def db_create(name, host = 'primary'):
	"""DB Create

	Creates a DB on the given host

	Arguments:
		name (str): The name of the DB to create
		host (str): The name of the host the DB will be on

	Returns:
		bool
	"""

	try:

		# Fetch the connection
		with _with(host) as oCon:

			# Create the DB
			dRes = r.db_create("%s%s" % (Record_Base.db_prepend(), name)).run(oCon)

			# If for some reason the DB wasn't created
			if 'dbs_created' not in dRes or not dRes['dbs_created']:
				return False

	# If the DB already exists
	except rerrors.ReqlOpFailedError:
		return True

	# Unknown runtime error
	except rerrors.RqlRuntimeError:
		return False

	# Return OK
	return True

def db_drop(name, host = 'primary'):
	"""DB Drop

	Drops a DB on the given host

	Arguments:
		name (str): The name of the DB to delete
		host (str): The name of the host the DB is on

	Returns:
		bool
	"""

	try:

		# Fetch the connection
		with _with(host) as oCon:

			# Delete the DB
			dRes = r.db_drop("%s%s" % (Record_Base.db_prepend(), name)).run(oCon)

			# If for some reason the DB wasn't dropped
			if 'dbs_dropped' not in dRes or not dRes['dbs_dropped']:
				return False

	# If the DB doesn't exist
	except rerrors.RqlRuntimeError:
		return False

	# Return OK
	return True

def db_prepend(pre = None):
	"""DB Prepend

	Gets or sets the global prefix for all DBs, useful for testing/development

	Arguments:
		pre (str): The prefix to store

	Returns:
		str|None
	"""
	return Record_Base.db_prepend(pre)

class Record(Record_Base.Record):
	"""Record

	Extends the base Record class
	"""

	@classmethod
	def add_changes(cls, _id, changes, customer={}):
		"""Add Changes

		Adds a record to the tables associated _changes table. Useful for
		Record types that can't handle multiple levels and have children
		tables that shouldn't be updated for every change in a single record

		Arguments:
			_id (mixed): The ID of the record the change is associated with
			changes (dict): The dictionary of changes to add
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			bool
		"""
		raise Exception('add_changes method not available in Record_ReDB')

	@classmethod
	def append(cls, _id, array, item, custom={}):
		"""Append

		Adds an item to a given array/list for a specific record

		Arguments:
			_id (mixed): The ID of the record to append to
			array (str): The name of the field with the array
			item (mixed): The value to append to the array
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			bool
		"""

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Append the org
			dRes = r \
				.db(dStruct['db']) \
				.table(dStruct['table']) \
				.get(_id) \
				.update({array: r.row[array].append(item)}) \
				.run(oCon)

			# Return True if a record was changed
			return dRes['replaced'] == 1

		# Return False
		return False

	@classmethod
	def config(cls):
		"""Config

		Returns the configuration data associated with the record type

		Returns:
			dict
		"""
		raise NotImplementedError('Must implement the "config" method')

	@classmethod
	def contains(cls, _id, array, item, custom={}):
		"""Contains

		Checks if a specific item exist inside a given array/list

		Arguments:
			_id (mixed): The ID of the record to check
			array (str): The name of the field with the array
			item (mixed): The value to check for in the array
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			bool
		"""

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Check if the org exists
			return r \
				.db(dStruct['db']) \
				.table(dStruct['table']) \
				.get(_id)[array] \
				.contains(item) \
				.run(oCon)

	@classmethod
	def count(cls, _id=None, index=None, filter=None, custom={}):
		"""Count

		Returns the number of records associated with index or filter

		Arguments:
			_ids (mixed): The ID to check
			index (str): Used as the index instead of the primary key
			filter (dict): Additional filter
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			bool
		"""

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# If an index is passed
		if index:

			# If the index doesn't exist
			if index not in dStruct['indexes']:
				raise KeyError('no index named "%s" in the structure' % index)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Create a cursor for all records
			oCur = r \
				.db(dStruct['db']) \
				.table(dStruct['table'])

			# If there's no primary key, we want all records
			if _id == None:
				pass

			# Else, if there's an index
			elif index:

				# If we recieved a dict for the primary key
				if isinstance(_id, dict):

					# Between two points
					if 'between' in _id:
						oCur = oCur.between(_id['between'][0], _id['between'][1], index=index, right_bound='closed')

					# Greater than
					elif 'gt' in _id:
						oCur = oCur.between(_id['gt'], r.maxval, index=index, left_bound='open')

					# Greater than or equal
					elif 'gte' in _id:
						oCur = oCur.between(_id['gte'], r.maxval, index=index)

					# Less than
					elif 'lt' in _id:
						oCur = oCur.between(r.minval, _id['lt'], index=index)

					# Less than or equal
					elif 'lte' in _id:
						oCur = oCur.between(r.minval, _id['lte'], index=index, right_bound='closed')

					# Invalid request
					else:
						raise ValueError('_id', _id)

				# If we received a tuple
				elif isinstance(_id, tuple):

					# Look for None values
					iNone = -1
					for i in range(len(_id)):
						if _id[i] is None:
							if iNone != -1:
								raise ValueError('_id', 'only one None per tuple')
							iNone = i

					# If there us a None in the tuple, assume between and
					#	replace them with the min and max
					if iNone > -1:
						idMax = list(_id)
						idMin = list(_id)
						idMax[iNone] = r.maxval
						idMin[iNone] = r.minval
						oCur = oCur.between(idMin, idMax, index=index)

					# No None, pass as is
					else:
						oCur = oCur.get_all(_id, index=index)

				# If we get a list, it's a complex index
				elif isinstance(_id, list):
					oCur = oCur.get_all(r.args(_id), index=index)

				# Else, pass as is
				else:
					oCur = oCur.get_all(_id, index=index)

			# If we are using the primary key
			else:

				# If we recieved a dict for the primary key
				if isinstance(_id, dict):

					# Between two points
					if 'between' in _id:
						oCur = oCur.between(_id['between'][0], _id['between'][1], right_bound='closed')

					# Greater than
					elif 'gt' in _id:
						oCur = oCur.between(_id['gt'], r.maxval, left_bound='open')

					# Greater than or equal
					elif 'gte' in _id:
						oCur = oCur.between(_id['gte'], r.maxval)

					# Less than
					elif 'lt' in _id:
						oCur = oCur.between(r.minval, _id['lt'])

					# Less than or equal
					elif 'lte' in _id:
						oCur = oCur.between(r.minval, _id['lte'], right_bound='closed')

					# Invalid request
					else:
						raise ValueError('_id', _id)

				# If we received a tuple
				elif isinstance(_id, tuple):

					# Look for None values
					iNone = -1
					for i in range(len(_id)):
						if _id[i] is None:
							if iNone != -1:
								raise ValueError('_id', 'only one None per tuple')
							iNone = i

					# If there us a None in the tuple, assume between and
					#	replace them with the min and max
					if iNone > -1:
						idMax = list(_id)
						idMin = list(_id)
						idMax[iNone] = r.maxval
						idMin[iNone] = r.minval
						oCur = oCur.between(idMin, idMax)

					# No None, pass as is
					else:
						bMulti = False
						oCur = oCur.get(_id)

				# If we got multiple primary keys, use get all
				elif isinstance(_id, list):
					oCur = oCur.get_all(*_id)

				# If we want one record, change the multi flag, and use get
				else:
					bMulti = False
					oCur = oCur.get(_id)

			# If we want to filter the data further
			if filter:

				# If we got a list
				if isinstance(filter, (list,tuple)):
					oCur = oCur.filter(*filter)
				else:
					oCur = oCur.filter(filter)

			# Run the request and return the count
			try:
				return oCur.count().run(oCon)

			# Catch operational errors
			except rerrors.ReqlOpFailedError as e:

				# An invalid index was passed
				if e.args[0][:5] == 'Index':
					raise KeyError('no index named "%s" in the table' % index)

				# Else, re-raise
				raise e

	def create(self, conflict='error', changes=None):
		"""Create

		Adds the record to the DB and returns the primary key

		Arguments:
			conflict (str): Must be one of 'error', 'replace', or 'update'
			changes (dict): Data needed to store a change record, is
				dependant on the 'changes' config value

		Returns:
			mixed|None
		"""

		# Make sure conflict arg is valid
		if conflict not in ('error', 'replace', 'update'):
			raise ValueError('conflict', conflict)

		# If the record requires revisions, make the first one
		if self._dStruct['revisions']:
			self._revision(True)

		# Get a connection to the host
		with _with(self._dStruct['host']) as oCon:

			# Create a new record
			dRes = r \
				.db(self._dStruct['db']) \
				.table(self._dStruct['table']) \
				.insert(self._dRecord, conflict=conflict) \
				.run(oCon)

			# If we got an error
			if dRes['errors']:

				# If it's a duplicate key
				if dRes['first_error'][0:21] == 'Duplicate primary key':
					raise Record_Base.DuplicateException(self._dStruct['primary'], self._dRecord[self._dStruct['primary']])

				# Else, it's an unknown error
				else:
					raise Exception(dRes['first_error'])

			# If the record was not inserted for some reason
			if dRes['inserted'] != 1 and dRes['replaced'] != 1:
				return None

			# Store the primary key if we didn't set it ourselves
			if self._dStruct['auto_primary'] and self._dStruct['primary'] not in self._dRecord:
				self._dRecord[self._dStruct['primary']] = dRes['generated_keys'][0]

			# If changes are required
			if self._dStruct['changes']:

				# Get the current timestamp
				iTime = int(time())

				# Create the changes record
				dChanges = {
					self._dStruct['primary']: self._dRecord[self._dStruct['primary']],
					"last": iTime,
					"items": {
						str(iTime): {
							"old": None,
							"new": "inserted"
						}
					}
				}

				# If Changes requires fields
				if isinstance(self._dStruct['changes'], list):

					# If they weren't passed
					if not isinstance(changes, dict):
						raise ValueError('changes')

					# Else, add the extra fields
					for k in self._dStruct['changes']:
						dChanges['items'][str(iTime)][k] = changes[k]

				# Create changes record
				dRes = r \
					.db(self._dStruct['db']) \
					.table("%s_changes" % self._dStruct['table']) \
					.insert(dChanges) \
					.run(oCon)

		# Return the primary key
		return self._dRecord[self._dStruct['primary']]

	@classmethod
	def create_many(cls, records, conflict='error', custom={}):
		"""Create Many

		Inserts multiple records at once, returning all their primary keys
		if auto_primary is true, else just returning the number of records
		inserted (or replaced if replace is set to True)

		Arguments:
			records (Record_ReDB.Record[]): A list of Record instances to insert
			conflict (str): Must be one of 'error', 'replace', or 'update'
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			uint
		"""

		# Make sure conflict arg is valid
		if conflict not in ('error', 'replace', 'update'):
			raise ValueError('conflict', conflict)

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# If changes are required
		if dStruct['changes']:
			raise RuntimeError('Tables with \'changes\' flag can\'t be inserted using create_many')

		# Initialise a list of raw records
		lRecords = []

		# Loop through the records
		for o in records:

			# If the record requires revisions
			if dStruct['revisions']:
				o._revision(True)

			# Add the record
			lRecords.append(o._dRecord)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Create a new record
			dRes = r \
				.db(dStruct['db']) \
				.table(dStruct['table']) \
				.insert(lRecords, conflict=conflict) \
				.run(oCon)

			# If we got an error
			if dRes['errors']:

				# If it's a duplicate key
				if dRes['first_error'][0:21] == 'Duplicate primary key':
					raise Record_Base.DuplicateException(dStruct['primary'])

				# Else, it's an unknown error
				else:
					raise Exception(dRes['first_error'])

			# If the record was not inserted for some reason
			if dRes['inserted'] == 0 and dRes['replaced'] == 0:
				return None

			# Go through each record and store the primary key if we didn't set
			#	it ourselves
			if dStruct['auto_primary']:
				for i in range(len(records)):
					records[i][dStruct['primary']] = dRes['generated_keys'][i]

				# Return the primary keys
				return dRes['generated_keys']

			# Else return how many records were replaced or inserted
			else:
				return dRes['inserted'] + dRes['replaced']

	def delete(self, changes=None):
		"""Delete

		Deletes the record represented by the instance

		Arguments:
			changes (dict): Data needed to store a change record, is
				dependant on the 'changes' config value

		Returns:
			bool
		"""

		# If the record lacks a primary key (never been created/inserted)
		if self._dStruct['primary'] not in self._dRecord:
			raise KeyError(self._dStruct['primary'])

		# Get a connection to the host
		with _with(self._dStruct['host']) as oCon:

			# Delete by primary key
			dRes = r \
				.db(self._dStruct['db']) \
				.table(self._dStruct['table']) \
				.get(self._dRecord[self._dStruct['primary']]) \
				.delete() \
				.run(oCon)

			# If there was an error
			if dRes['deleted'] != 1:
				return False

			# If changes are required
			if self._dStruct['changes']:

				# Get the current timestamp
				iTime = int(time())

				# Create the changes record
				dChanges = {
					"last": iTime,
					"items": {
						str(iTime): {
							"old": self._dRecord,
							"new": None
						}
					}
				}

				# If Changes requires fields
				if isinstance(self._dStruct['changes'], list):

					# If they weren't passed
					if not isinstance(changes, dict):
						raise ValueError('changes')

					# Else, add the extra fields
					for k in self._dStruct['changes']:
						dChanges['items'][str(iTime)][k] = changes[k]

				# Update the changes
				dRes = r \
					.db(self._dStruct['db']) \
					.table("%s_changes" % self._dStruct['table']) \
					.get(self._dRecord[self._dStruct['primary']]) \
					.update(dChanges) \
					.run(oCon)

			# Remove the primary key value so we can't delete again or save
			del self._dRecord[self._dStruct['primary']]

		# Return OK
		return True

	@classmethod
	def delete_get(cls, _id=None, index=None, custom={}):
		"""Delete Get

		Deletes one or many records by primary key or index and returns how many
		were found/deleted

		Arguments:
			_id (mixed|mixed[]): The primary key(s) to delete or None for all
				records
			index (str): Used as the index instead of the primary key
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Return:
			int
		"""

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# If changes are required
		if dStruct['changes']:
			raise RuntimeError('Tables with \'changes\' flag can\'t be deleted using delete_get')

		# If an index is passed
		if index:

			# If the index doesn't exist
			if index not in dStruct['indexes']:
				raise KeyError('no index named "%s" in the tree' % index)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Create a cursor for all records
			oCur = r \
				.db(dStruct['db']) \
				.table(dStruct['table']) \

			# If there's no primary key, we want all records
			if _id is None:
				pass

			# If an index was passed, use get all
			elif index:

				# If we recieved a dict for the primary key
				if isinstance(_id, dict):

					# Between two points
					if 'between' in _id:
						oCur = oCur.between(_id['between'][0], _id['between'][1], index=index, right_bound='closed')

					# Greater than
					elif 'gt' in _id:
						oCur = oCur.between(_id['gt'], r.maxval, index=index, left_bound='open')

					# Greater than or equal
					elif 'gte' in _id:
						oCur = oCur.between(_id['gte'], r.maxval, index=index)

					# Less than
					elif 'lt' in _id:
						oCur = oCur.between(r.minval, _id['lt'], index=index)

					# Less than or equal
					elif 'lte' in _id:
						oCur = oCur.between(r.minval, _id['lte'], index=index, right_bound='closed')

					# Invalid request
					else:
						raise ValueError('_id', _id)

				# If we received a tuple
				elif isinstance(_id, tuple):

					# Look for None values
					iNone = -1
					for i in range(len(_id)):
						if _id[i] is None:
							if iNone != -1:
								raise ValueError('_id', 'only one None per tuple')
							iNone = i

					# If there us a None in the tuple, assume between and
					#	replace them with the min and max
					if iNone > -1:
						idMax = list(_id)
						idMin = list(_id)
						idMax[iNone] = r.maxval
						idMin[iNone] = r.minval
						oCur = oCur.between(idMin, idMax, index=index)

					# No None, pass as is
					else:
						oCur = oCur.get_all(_id, index=index)

			# If we are using the primary key
			else:

				# If we got multiple primary keys, use get all
				if isinstance(_id, (tuple,list)):
					oCur = oCur.get_all(*_id)

				# Else, use get for one record
				else:
					oCur = oCur.get(_id)

			# Delete the record(s)
			try:
				dRes = oCur.delete().run(oCon)

			# Catch operational errors
			except rerrors.ReqlOpFailedError as e:

				# An invalid index was passed
				if e.args[0][:5] == 'Index':
					raise KeyError('no index named "%s" in the table' % index)

				# Else, re-raise
				raise e

			# Return how many were deleted
			return dRes['deleted']

	@classmethod
	def exists(cls, _id, index=None, custom={}):
		"""Exists

		Returns the ID (primary key) of the record for the specified ID or
		unique index value found, else False if no record is found

		Arguments:
			_id (mixed): The primary key to check
			index (str): Used as the index instead of the primary key
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			bool
		"""

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# Use the get method to avoid duplicate code and check if anything was
		#	returned
		dRecord = cls.get(_id, index=index, raw=[dStruct['primary']], custom=custom)
		if not dRecord:
			return False

		# If anything was returned, return the primary key
		return dRecord[dStruct['primary']]

	# filter static method
	@classmethod
	def filter(cls, fields, raw=None, distinct=False, orderby=None, limit=None, custom={}):
		"""Filter

		Finds records based on the specific fields and values passed

		Arguments:
			fields (dict): A dictionary of field names to the values they
				should match
			raw (bool|list): Return raw data (dict) for all or a set list of
				fields
			distinct (bool): Only return distinct data
			orderby (str|str[]): A field or fields to order the results by
			limit (int|tuple): The limit and possible starting point
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			Record[]|dict[]
		"""

		# By default we will return multiple records
		bMulti = True

		# By default there is no limit
		iLimitParams = 0

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Use the passed fields to generate the request
			oCur = r \
				.db(dStruct['db']) \
				.table(dStruct['table']) \
				.filter(fields)

			# If we only want specific fields
			if isinstance(raw, (tuple,list)):
				oCur = oCur.pluck(*raw).default(None)

			# If distinct is requested
			if distinct:
				if isinstance(distinct, str):
					oCur = oCur.distinct(index=distinct)
				else:
					oCur = oCur.distinct()

			# If an order by arg was passed
			if orderby is not None:

				# If we need to orger by multiple fields
				if isinstance(orderby, (tuple,list)):
					oCur = oCur.order_by(*orderby)

				# If we need to order by
				elif isinstance(orderby, str):
					oCur = oCur.order_by(orderby)

				# Else we got something invalid
				else:
					raise ValueError('orderby', orderby)

			# If we recieved a limit
			if limit is not None:

				# If we got an int, we are only limiting
				if isinstance(limit, int):
					iLimitParams = 1
					oCur = oCur.limit(limit)

				# If we got a tuple
				elif isinstance(limit, (list,tuple)):
					iLimitParams = 2
					oCur = oCur.skip(limit[0]).limit(limit[1])

				# Else we got something invalid
				else:
					raise ValueError('limit', limit)

			# Run the request
			itRes = oCur.run(oCon)

			# If there's no data, return None or an empty list
			if not itRes:
				if bMulti: return []
				else: return None

			# If we are expecting a single record
			if (iLimitParams == 1 and limit == 1) or \
				(iLimitParams == 2 and limit[1] == 1):

				# If we got a list
				if isinstance(itRes, (tuple,list)):

					# If there's no data
					if not len(itRes):
						return None

					# Store the row
					dRow = itRes[0]

				# Else it's a cursor
				else:

					# Try to get one row
					try:
						dRow = itRes.next()
					except rnet.DefaultCursorEmpty as e:
						return None

				# If it's raw, don't instantiate it
				return (raw and dRow or cls(dRow, custom))

			# If Raw requested, return as is
			if raw:
				return [d for d in itRes]

			# Else create instances for each
			else:
				return [cls(d, custom) for d in itRes]

	@classmethod
	def generate_config(cls, tree, special='rethinkdb', override=None):
		"""Generate Config

		Generates record specific config based on the Format-OC tree passed

		Arguments:
			tree (FormatOC.Tree): the tree associated with the record type
			special (str): The special section used to identify the child info
			override (dict): Used to override any data from the tree

		Returns:
			dict
		"""

		# Call the parent
		return super().generate_config(tree, special, override)

	@classmethod
	def get(cls, _id=None, index=None, filter=None, match=None, raw=None, distinct=False, orderby=None, limit=None, custom={}):
		"""Get

		Returns records by primary key or index, can also be given an extra filter

		Arguments:
			_id (str|str[]): The primary key(s) to fetch from the table
			index (str): Index to use instead of primary key
			filter (dict): Additional filter
			match (tuple): Name/Match filter
			raw (bool|list): Return raw data (dict) for all or a set list of
				fields
			distinct (bool): Only return distinct data
			orderby (str|str[]): A field or fields to order the results by
			limit (int|tuple): The limit and possible starting point
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			Record|Record[]|dict|dict[]
		"""

		# By default we will return multiple records
		bMulti = True

		# By default there is no limit
		iLimitParams = 0

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# If an index is passed
		if index:

			# If the index doesn't exist
			if index not in dStruct['indexes']:
				raise KeyError('no index named "%s" in the structure' % index)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Create a cursor for all records
			oCur = r \
				.db(dStruct['db']) \
				.table(dStruct['table'])

			# If there's no primary key, we want all records
			if _id == None:
				pass

			# Else, if there's an index
			elif index:

				# If we recieved a dict for the primary key
				if isinstance(_id, dict):

					# Between two points
					if 'between' in _id:
						oCur = oCur.between(_id['between'][0], _id['between'][1], index=index, right_bound='closed')

					# Greater than
					elif 'gt' in _id:
						oCur = oCur.between(_id['gt'], r.maxval, index=index, left_bound='open')

					# Greater than or equal
					elif 'gte' in _id:
						oCur = oCur.between(_id['gte'], r.maxval, index=index)

					# Less than
					elif 'lt' in _id:
						oCur = oCur.between(r.minval, _id['lt'], index=index)

					# Less than or equal
					elif 'lte' in _id:
						oCur = oCur.between(r.minval, _id['lte'], index=index, right_bound='closed')

					# Invalid request
					else:
						raise ValueError('_id', _id)

				# If we received a tuple
				elif isinstance(_id, tuple):

					# Look for None values
					iNone = -1
					for i in range(len(_id)):
						if _id[i] is None:
							if iNone != -1:
								raise ValueError('_id', 'only one None per tuple')
							iNone = i

					# If there us a None in the tuple, assume between and
					#	replace them with the min and max
					if iNone > -1:
						idMax = list(_id)
						idMin = list(_id)
						idMax[iNone] = r.maxval
						idMin[iNone] = r.minval
						oCur = oCur.between(idMin, idMax, index=index)

					# No None, pass as is
					else:
						oCur = oCur.get_all(_id, index=index)

				# If we get a list, it's a complex index
				elif isinstance(_id, list):
					oCur = oCur.get_all(r.args(_id), index=index)

				# Else, pass as is
				else:
					oCur = oCur.get_all(_id, index=index)

			# If we are using the primary key
			else:

				# If we recieved a dict for the primary key
				if isinstance(_id, dict):

					# Between two points
					if 'between' in _id:
						oCur = oCur.between(_id['between'][0], _id['between'][1], right_bound='closed')

					# Greater than
					elif 'gt' in _id:
						oCur = oCur.between(_id['gt'], r.maxval, left_bound='open')

					# Greater than or equal
					elif 'gte' in _id:
						oCur = oCur.between(_id['gte'], r.maxval)

					# Less than
					elif 'lt' in _id:
						oCur = oCur.between(r.minval, _id['lt'])

					# Less than or equal
					elif 'lte' in _id:
						oCur = oCur.between(r.minval, _id['lte'], right_bound='closed')

					# Invalid request
					else:
						raise ValueError('_id', _id)

				# If we received a tuple
				elif isinstance(_id, tuple):

					# Look for None values
					iNone = -1
					for i in range(len(_id)):
						if _id[i] is None:
							if iNone != -1:
								raise ValueError('_id', 'only one None per tuple')
							iNone = i

					# If there us a None in the tuple, assume between and
					#	replace them with the min and max
					if iNone > -1:
						idMax = list(_id)
						idMin = list(_id)
						idMax[iNone] = r.maxval
						idMin[iNone] = r.minval
						oCur = oCur.between(idMin, idMax)

					# No None, pass as is
					else:
						bMulti = False
						oCur = oCur.get(_id)

				# If we got multiple primary keys, use get all
				elif isinstance(_id, list):
					oCur = oCur.get_all(*_id)

				# If we want one record, change the multi flag, and use get
				else:
					bMulti = False
					oCur = oCur.get(_id)

			# If we want to filter the data further
			if filter:

				# If we got a list
				if isinstance(filter, (list,tuple)):
					oCur = oCur.filter(*filter)
				else:
					oCur = oCur.filter(filter)

			# If we want to filter by a match
			if match:
				oCur = oCur.filter(lambda doc: doc[match[0]].match(match[1]))

			# If we only want specific fields
			if isinstance(raw, (tuple,list)):
				oCur = oCur.pluck(*raw).default(None)

			# If distinct is requested
			if distinct:
				if isinstance(distinct, str):
					oCur = oCur.distinct(index=distinct)
				else:
					oCur = oCur.distinct()

			# If an order by arg was passed
			if orderby is not None:

				# If we need to order by multiple fields
				if isinstance(orderby, (tuple,list)):
					for i in range(len(orderby)):
						if(orderby[i][0] == '!'):
							orderby[i] = r.desc(orderby[i][1:])
					oCur = oCur.order_by(*orderby)

				# If we need to order by
				elif isinstance(orderby, str):
					if(orderby[0] == '!'):
						orderby = r.desc(orderby[1:])
					oCur = oCur.order_by(orderby)

				# Else we got something invalid
				else:
					raise ValueError('orderby', orderby)

			# If we recieved a limit
			if limit is not None:

				# If we got an int, we are only limiting
				if isinstance(limit, int):
					iLimitParams = 1
					oCur = oCur.limit(limit)

					# If we only want one record
					if limit == 1:
						bMulti = False

				# If we got a tuple
				elif isinstance(limit, (list,tuple)):
					iLimitParams = 2
					oCur = oCur.skip(limit[0]).limit(limit[1])

					# If we only want one record
					if limit[1] == 1:
						bMulti = False

				# Else we got something invalid
				else:
					raise ValueError('limit', limit)

			# Run the request
			try:
				itRes = oCur.run(oCon)

			# Catch operational errors
			except rerrors.ReqlOpFailedError as e:

				# An invalid index was passed
				if e.args[0][:5] == 'Index':
					raise KeyError('no index named "%s" in the table' % index)

				# Else, re-raise
				raise e

			# If there's no data, return None or an empty list
			if not itRes:
				if bMulti: return []
				else: return None

			# If we are expecting a single record
			if (iLimitParams == 1 and limit == 1) or \
				(iLimitParams == 2 and limit[1] == 1):

				# If we got a list
				if isinstance(itRes, (tuple,list)):

					# If there's no data
					if not len(itRes):
						return None

					# Store the row
					dRow = itRes[0]

				# Else it's a cursor
				else:

					# Try to get one row
					try:
						dRow = itRes.next()
					except rnet.DefaultCursorEmpty as e:
						return None

				# If it's raw, don't instantiate it
				return (raw and dRow or cls(dRow, custom))

			# If the multi flag is still set
			if bMulti:

				# If Raw requested, return as is
				if raw:
					return [d for d in itRes]

				# Else create instances for each
				else:
					return [cls(d, custom) for d in itRes]

			# Else, one record requested
			else:
				return raw and itRes or cls(itRes, custom)

	@classmethod
	def get_changes(cls, _id, orderby=None, custom={}):
		"""Get Changes

		Returns the changes record associated with the primary record and table.
		Used by Record types that have the 'changes' flag set

		Arguments:
			_id (mixed): The of the primary record to fetch changes for
			orderby (str|str[]): A field or fields to order the results by
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			dict
		"""

		# By default we will return multiple records
		bMulti = True

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Create a cursor for all records
			oCur = r \
				.db(dStruct['db']) \
				.table("%s_changes" % dStruct['table'])

			# If there's no primary key, we want all records
			if _id == None:
				pass

			# Else, we want specific keys
			else:

				# If we got multiple primary keys, use get all
				if isinstance(_id, (tuple,list)):
					oCur = oCur.get_all(*_id)

				# If we want one record, change the multi flag, and use get
				else:
					bMulti = False
					oCur = oCur.get(_id)

			# If an order by arg was passed
			if orderby is not None:

				# If we need to orger by multiple fields
				if isinstance(orderby, (tuple,list)):
					oCur = oCur.order_by(*orderby)

				# If we need to order by
				elif isinstance(orderby, str):
					oCur = oCur.order_by(orderby)

				# Else we got something invalid
				else:
					raise ValueError('orderby', orderby)

			# Run the request
			itRes = oCur.run(oCon)

			# If there's no data, return None or an empty list
			if not itRes:
				if bMulti: return []
				else: return None

			# If the multi flag is still set
			if bMulti:

				# Return all records
				return [d for d in itRes]

			# Else, one record requested
			else:
				return itRes

	@classmethod
	def remove(cls, _id, array, index, custom={}):
		"""Remove

		Removes an item from a given array/list for a specific record

		Arguments:
			_id (mixed): The ID of the record to remove from
			array (str): The name of the field with the array
			index (uint): The index of the array to remove
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			bool
		"""

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Append the org
			dRes = r \
				.db(dStruct['db']) \
				.table(dStruct['table']) \
				.get(_id) \
				.update({array: r.row[array].delete_at(index)}) \
				.run(oCon)

			# Return True if a record was changed
			return dRes['replaced'] == 1

		# Return False
		return False

	def save(self, replace=False, changes=None):
		"""Save

		Updates the record in the DB and returns true if anything has changed,
		or a new revision number of the record is revisionable

		Arguments:
			replace (bool): If true, replace all fields instead of updating
			changes (dict): Data needed to store a change record, is
				dependant on the 'changes' config value

		Returns:
			bool
		"""

		# If no fields have been changed, nothing to do
		if not self._dChanged:
			return False

		# If there is no primary key in the record
		if self._dStruct['primary'] not in self._dRecord:
			raise KeyError(self._dStruct['primary'])

		# Start by fetching the record
		oCur = r \
			.db(self._dStruct['db']) \
			.table(self._dStruct['table']) \
			.get(self._dRecord[self._dStruct['primary']])

		# If a replace was requested, or all fields have been changed
		if replace or (isinstance(self._dChanged, bool) and self._dChanged):
			oCur = oCur.replace(self._dRecord)

		# Else we are updating
		else:
			oCur = oCur.update({k:self._dRecord[k] for k in self._dChanged})

		# Get a connection to the host
		with _with(self._dStruct['host']) as oCon:

			# If revisions are required
			if self._dStruct['revisions']:

				# Store the old revision
				sRev = self._dRecord[self._dStruct['rev_field']]

				# If updating the revision fails
				if not self._revision():
					return False

				# Use the primary key to fetch the record and return the rev
				dRecord = r \
					.db(self._dStruct['db']) \
					.table(self._dStruct['table']) \
					.get(self._dRecord[self._dStruct['primary']]) \
					.pluck([self._dStruct['rev_field']]) \
					.default(None) \
					.run(oCon)

				# If there's no such record
				if not dRecord:
					return False

				# If it is found, but the revisions don't match up
				if dRecord[self._dStruct['rev_field']] != sRev:
					raise Record_Base.RevisionException(self._dRecord[self._dStruct['primary']])

			# Update the record
			dRes = oCur.run(oCon)

			# If we got an error
			if dRes['errors']:

				# If it's a duplicate key
				if dRes['first_error'][0:21] == 'Duplicate primary key':
					raise Record_Base.DuplicateException(self._dStruct['primary'], self._dRecord[self._dStruct['primary']])

				# Else, it's an unknown error
				else:
					raise Exception(dRes['first_error'])

			# If the record wasn't updated for some reason
			if dRes['replaced'] != 1:
				return False

			# If changes are required
			if self._dStruct['changes'] and changes != False:

				# Get the current timestamp
				iTime = int(time())

				# Create the changes record
				dChanges = {
					"last": iTime,
					"items": {
						str(iTime): self.generate_changes(
							self._dOldRecord,
							self._dRecord
						)
					}
				}

				# If Changes requires fields
				if isinstance(self._dStruct['changes'], list):

					# If they weren't passed
					if not isinstance(changes, dict):
						raise ValueError('changes')

					# Else, add the extra fields
					for k in self._dStruct['changes']:
						dChanges['items'][str(iTime)][k] = changes[k]

				# Update the changes
				dRes = r \
					.db(self._dStruct['db']) \
					.table("%s_changes" % self._dStruct['table']) \
					.get(self._dRecord[self._dStruct['primary']]) \
					.update(dChanges) \
					.run(oCon)

				# Reset the old record
				self._dOldRecord = None

		# Clear the changed fields flags
		self._dChanged = {}

		# Return OK
		return True

	@classmethod
	def table_create(cls, custom={}):
		"""Table Create

		Creates the record's table/collection/etc in the DB

		Arguments:
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			bool
		"""

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Try to create the table
			try:
				dRes = r \
					.db(dStruct['db']) \
					.table_create(dStruct['table'], primary_key=dStruct['primary']) \
					.run(oCon)

				# If for some reason the table wasn't created
				if 'tables_created' not in dRes or not dRes['tables_created']:
					return False

				# If indexes are specified for the record
				if dStruct['indexes']:

					# Loop through the indexes to get the name and fields
					for sName,mFields in dStruct['indexes'].items():

						# Start with the table
						oCur = r \
							.db(dStruct['db']) \
							.table(dStruct['table'])

						# No field means the name is the field
						if mFields is None:
							dRes = oCur.index_create(sName).run(oCon)

						# One field
						elif isinstance(mFields, str):
							dRes = oCur.index_create(sName, r.row[mFields]).run(oCon)

						# A list of fields
						elif isinstance(mFields, (tuple,list)):
							lFields = []
							for sField in mFields:
								lFields.append(r.row[sField])
							dRes = oCur.index_create(sName, lFields).run(oCon)

						# Invalid field value
						else:
							raise ValueError('indexes', str(mFields))

				# If changes are required
				if dStruct['changes']:

					# Try to create the table
					dRes = r \
						.db(dStruct['db']) \
						.table_create("%s_changes" % dStruct['table'], primary_key=dStruct['primary']) \
						.run(oCon)

			# The table already exists
			except rerrors.RqlRuntimeError as e:
				print(str(e))
				return False

		# Return OK
		return True

	@classmethod
	def table_drop(cls, custom={}):
		"""Table Drop

		Deletes the record's table/collection/etc in the DB

		Arguments:
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			bool
		"""

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Try to drop the table
			try:
				dRes = r \
					.db(dStruct['db']) \
					.table_drop(dStruct['table']) \
					.run(oCon)

				# If for some reason the table wasn't dropped
				if 'tables_dropped' not in dRes or not dRes['tables_dropped']:
					return False

				# If changes are required
				if dStruct['changes']:

					# Try to delete the table
					dRes = r \
						.db(dStruct['db']) \
						.table_drop("%s_changes" % dStruct['table']) \
						.run(oCon)

			# If there's no such table
			except rerrors.RqlRuntimeError:
				return False

		# Return OK
		return True

	@classmethod
	def update_field(cls, field, value, _id=None, index=None, filter=None, custom={}):
		"""Updated Field

		Updates a specific field to the value for an ID, many IDs, or the entire
		table

		Arguments:
			field (str): The name of the field to update
			value (mixed): The value to set the field to
			_id (mixed): Optional ID(s) to filter by
			index (str): Optional name of the index to use instead of primary
			filter (dict): Optional filter list to decide what records get updated
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			uint -- Number of records altered
		"""

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# If the field doesn't exist
		if field not in dStruct['tree']:
			raise KeyError('field', field)

		# If an index is passed
		if index:

			# If the index doesn't exist
			if index not in dStruct['indexes']:
				raise KeyError('no index named "%s" in the structure' % index)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Create a cursor for all records
			oCur = r \
				.db(dStruct['db']) \
				.table(dStruct['table'])

			# If there's no primary key, nothing else to do with indexes or IDs
			if _id == None:
				pass

			# Else, if there's an index
			elif index:

				# If we recieved a dict for the primary key
				if isinstance(_id, dict):

					# Between two points
					if 'between' in _id:
						oCur = oCur.between(_id['between'][0], _id['between'][1], index=index, right_bound='closed')

					# Greater than
					elif 'gt' in _id:
						oCur = oCur.between(_id['gt'], r.maxval, index=index, left_bound='open')

					# Greater than or equal
					elif 'gte' in _id:
						oCur = oCur.between(_id['gte'], r.maxval, index=index)

					# Less than
					elif 'lt' in _id:
						oCur = oCur.between(r.minval, _id['lt'], index=index)

					# Less than or equal
					elif 'lte' in _id:
						oCur = oCur.between(r.minval, _id['lte'], index=index, right_bound='closed')

					# Invalid request
					else:
						raise ValueError('_id', _id)

				# If we received a tuple
				elif isinstance(_id, tuple):

					# Look for None values
					iNone = -1
					for i in range(len(_id)):
						if _id[i] is None:
							if iNone != -1:
								raise ValueError('_id', 'only one None per tuple')
							iNone = i

					# If there us a None in the tuple, assume between and
					#	replace them with the min and max
					if iNone > -1:
						idMax = list(_id)
						idMin = list(_id)
						idMax[iNone] = r.maxval
						idMin[iNone] = r.minval
						oCur = oCur.between(idMin, idMax, index=index)

					# No None, pass as is
					else:
						oCur = oCur.get_all(_id, index=index)

				# If we get a list, it's a complex index
				elif isinstance(_id, list):
					oCur = oCur.get_all(r.args(_id), index=index)

				# Else, pass as is
				else:
					oCur = oCur.get_all(_id, index=index)

			# If we are using the primary key
			else:

				# If we recieved a dict for the primary key
				if isinstance(_id, dict):

					# Between two points
					if 'between' in _id:
						oCur = oCur.between(_id['between'][0], _id['between'][1], right_bound='closed')

					# Greater than
					elif 'gt' in _id:
						oCur = oCur.between(_id['gt'], r.maxval, left_bound='open')

					# Greater than or equal
					elif 'gte' in _id:
						oCur = oCur.between(_id['gte'], r.maxval)

					# Less than
					elif 'lt' in _id:
						oCur = oCur.between(r.minval, _id['lt'])

					# Less than or equal
					elif 'lte' in _id:
						oCur = oCur.between(r.minval, _id['lte'], right_bound='closed')

					# Invalid request
					else:
						raise ValueError('_id', _id)

				# If we got multiple primary keys, use get all
				elif isinstance(_id, (tuple,list)):
					oCur = oCur.get_all(*_id)

				# If we want one record, change the multi flag, and use get
				else:
					oCur = oCur.get(_id)

			# If we want to filter the data further
			if filter:
				oCur = oCur.filter(filter)

			# Add the update and run the request
			try:
				dRes = oCur.update({field: value}).run(oCon)

			# Catch operational errors
			except rerrors.ReqlOpFailedError as e:

				# An invalid index was passed
				if e.args[0][:5] == 'Index':
					raise KeyError('no index named "%s" in the table' % index)

				# Else, re-raise
				raise e

			# Return the number of rows changed
			return dRes['replaced']

	@classmethod
	def uuid(cls, custom={}):
		"""UUID

		Returns a universal unique ID

		Arguments:
			custom (dict): Custom Host and DB info
				'host' the name of the host to get/set data on
				'append' optional postfix for dynamic DBs

		Returns:
			str
		"""

		# Fetch the record structure
		dStruct = cls.struct(custom)

		# Get a connection to the host
		with _with(dStruct['host']) as oCon:

			# Get the UUID
			return r.uuid().run(oCon)

# Register the module with the Base
Record_Base.register_type('rethinkdb', sys.modules[__name__])



# coding=utf8
""" Environment

Shared functionality to deal with environment data
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2022-08-29"

def get_client_ip(environ):
	"""Get Client IP

	Returns the IP of the client based on all the environment data passed to
	the current webserver request

	Arguments:
		environ (dict): A dictionary of environment variables

	Returns:
		str
	"""

	# Init return var
	sIP	= '0.0.0.0'

	# Check common environment variables
	if 'HTTP_CLIENT_IP' in environ:
		sIP = environ['HTTP_CLIENT_IP']
	elif 'HTTP_X_CLIENTIP' in environ:
		sIP = environ['HTTP_X_CLIENTIP']
	elif 'HTTP_X_FORWARDED_FOR' in environ:
		sIP = environ['HTTP_X_FORWARDED_FOR']
	elif 'HTTP_X_RN_XFF' in environ:
		sIP = environ['HTTP_X_RN_XFF']
	elif 'REMOTE_ADDR' in environ:
		sIP = environ['REMOTE_ADDR']

	# If there's multiple IPs
	if sIP.find(','):
		lIPs = sIP.split(',')
		sIP = lIPs[-1].strip()

	# Return the IP
	return sIP



# coding=utf8
""" Session Module

Handles internal sessions shared across microservices
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-11"

# Python imports
import uuid

# Pip imports
from redis import StrictRedis

# Module imports
from . import JSON, StrHelper

# Open redis connection
_moRedis = None
_muiExpire = 86400

def create(id = None, expires=None):
	"""Create

	Returns a brand new session using the ID given, if no ID is passed, one is
	generated

	Arguments:
		id (str): The ID to use for the session
		expires (uint): A specific expiry in seconds to override the global one
						for this session

	Returns:
		Session
	"""

	# Init the data
	dData = {}

	# If we have an expires time
	if expires:
		dData['__expire'] = expires

	# Create a new Session using a UUID as the id
	return _Session(id and id or uuid.uuid4().hex, dData)

def init(conf, expire=0):
	"""Init

	Initialises the module

	Arguments:
		conf (dict): The necessary Redis config
		expire (uint): Length in seconds for the session to remain active

	Returns:
		None
	"""

	# Pull in the module variable
	global _moRedis, _muiExpire

	# Create the Redis connection
	_moRedis = StrictRedis(**conf)

	# Store the expire time
	_muiExpire = expire

def load(id):
	"""Load

	Loads an existing session from the cache

	Arguments:
		id (str): The unique id of an existing session

	Returns:
		Session
	"""

	# Fetch from Redis
	s = _moRedis.get(id)

	# If there's no session or it expired
	if s == None: return None

	# Make sure we have a string, not a set of bytes
	try: s = s.decode()
	except (UnicodeDecodeError, AttributeError): pass

	# Create a new instance with the decoded data
	return _Session(id, JSON.decode(s))

class _Session(object):
	"""Session

	A wrapper for the session data
	"""

	def __init__(self, id, data={}):
		"""Constructor

		Intialises the instance, which is just setting up the dict

		Arguments:
			id (str): The ID of the session
			data (dict): The data in the session

		Returns:
			Session
		"""
		self.__id = id
		self.__dStore = data

	def __contains__(self, key):
		"""__contains__

		True if the key exists in the session

		Arguments:
			key (str): The field to check for

		Returns:
			bool
		"""
		return key in self.__dStore

	def __delitem__(self, key):
		"""__delete__

		Removes a key from a session

		Arguments:
			key (str): The key to remove

		Returns:
			None
		"""
		del self.__dStore[key]

	def __getitem__(self, key):
		"""__getitem__

		Returns the given key

		Arguments:
			key (str): The key to return

		Returns:
			mixed
		"""
		return self.__dStore[key]

	def __iter__(self):
		"""__iter__

		Returns an iterator for the internal dict

		Returns:
			iterator
		"""
		return iter(self.__dStore)

	def __len__(self):
		"""__len__

		Return the length of the internal dict

		Returns:
			uint
		"""
		return len(self.__dStore)

	def __setitem__(self, key, value):
		"""__setitem__

		Sets the given key

		Arguments:
			key (str): The key to set
			value (mixed): The value for the key

		Returns:
			None
		"""
		self.__dStore[key] = value

	def __str__(self):
		"""__str__

		Returns a string representation of the internal dict

		Returns:
			str
		"""
		return str(self.__dStore)

	def close(self):
		"""Close

		Deletes the session from the cache

		Returns:
			None
		"""
		_moRedis.delete(self.__id)

	def extend(self):
		"""Extend

		Keep the session alive by extending it's expire time by the internally
		set expirty, or else by the global one set for the module

		Returns:
			None
		"""

		# Use internal time if we have one, else use the global
		iExpire = '__expire' in self.__dStore and \
					self.__dStore['__expire'] or \
					_muiExpire

		# If the expire time is 0, do nothing
		if iExpire == 0:
			return

		# Extend the session in Redis
		_moRedis.expire(self.__id, iExpire)

	def id(self):
		"""ID

		Returns the ID of the session

		Returns:
			str
		"""
		return self.__id

	def save(self):
		"""Save

		Saves the current session data in the cache

		Returns:
			None
		"""

		# Use internal time if we have one, else use the global
		iExpire = '__expire' in self.__dStore and self.__dStore['__expire'] or _muiExpire

		# If we have no expire time, set forever
		if iExpire == 0:
			_moRedis.set(self.__id, JSON.encode(self.__dStore))

		# Else, set to expire
		else:
			_moRedis.setex(self.__id, _muiExpire, JSON.encode(self.__dStore))



from setuptools import setup

with open('README.md', 'r') as oF:
	long_description=oF.read()

setup(
	name='rest-oc',
	version='1.2.8',
	description='Rest-OC is a library of python 3 modules for rapidly setting up REST microservices.',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://ouroboroscoding.com/rest-oc/',
	project_urls={
		'Documentation': 'https://ouroboroscoding.com/rest-oc/',
		'Source': 'https://github.com/ouroboroscoding/rest-oc-python',
		'Tracker': 'https://github.com/ouroboroscoding/rest-oc-python/issues'
	},
	keywords=['rest','microservices'],
	author='Chris Nasr - Ouroboros Coding Inc.',
	author_email='chris@ouroboroscoding.com',
	license='MIT',
	packages=['RestOC'],
	python_requires='>=3.10',
	install_requires=[
		'arrow>=1.2.3,<1.3',
		'bottle>=0.12.25,<0.13',
		'format-oc>=1.6.2,<1.7',
		'gunicorn>=20.1.0,<20.2',
		'Jinja2>=3.1.2,<3.2',
		'json-fix>=0.5.2,<0.6',
		'markupsafe>=2.0.1,<2.1',
		'pdfkit>=1.0.0,<1.1',
		'piexif>=1.1.3,<1.2',
		'Pillow>=9.4.0,<9.5',
		'PyMySQL>=1.0.2,<1.1',
		'redis>=4.5.1,<4.6',
		'requests>=2.28.2,<2.29',
		'rethinkdb>=2.4.9,<2.5'
	],
	zip_safe=True
)


# coding=utf8
"""JSON

Wrapper for Python json module which handles custom types
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2016-01-07"

# Import python modules
import json
from datetime import datetime
from decimal import Decimal
import os

# decode function
def decode(s):
	"""Decode

	Handles decoding JSON, as a string, into objects/values

	Args:
		s (str): The JSON to decode

	Returns:
		mixed
	"""
	return json.loads(s, parse_float=Decimal)

# decodef function
def decodef(f):
	"""Decode File

	Handles decoding JSON, from a file, into objects/values

	Args:
		f (file-like object): An instance of a file object with read/write
			methods

	Returns:
		mixed
	"""
	return json.load(f, parse_float=Decimal)

# encode function
def encode(o, indent=None):
	"""Encode

	Handles encoding objects/values into a JSON string

	Args:
		o (mixed): The object or value to encode

	Returns:
		str
	"""
	return json.dumps(o, cls=CEncoder, indent=indent)

# encodef function
def encodef(o, f, indent=None):
	"""Encode File

	Handles encoding objects/values into JSON and storing them in the given file

	Args:
		o (mixed): The object or value to encode
		f (file-like object): An instance of a file object with read/write
			methods

	Returns:
		None
	"""
	return json.dump(o, f, cls=CEncoder, indent=indent)

# load function
def load(filepath):
	"""Load

	Loads a data structure from a JSON file given a full or relative path to it

	Args:
		filepath (str): The path to the file to load

	Returns:
		mixed
	"""

	# If ~ is present
	if '~' in filepath:
		filepath = os.path.expanduser(filepath)

	# Load the file
	with open(filepath, 'r') as oFile:

		# Convert it to a python variable and return it
		return decode(oFile.read())

# store function
def store(data, filepath, indent=None):
	"""Store

	Converts an object/value into JSON and stores it in the file path given

	Args:
		filepath (str): The full or relative path to the file

	Returns:
		None
	"""

	# If ~ is present
	if '~' in filepath:
		filepath = os.path.expanduser(filepath)

	# Open a file to write the data
	with open(filepath, 'w') as oFile:

		# Write the JSON to the file
		oFile.write(encode(data, indent))

# CEncoder class
class CEncoder(json.JSONEncoder):
	"""Encoder

	Handles encoding types the default JSON encoder can't handle
	"""

	# default method
	def default(self, obj):
		"""Default

		Called when the regular Encoder can't figure out what to do with the
		type

		Args:
			obj (mixed): An unknown object or value that needs to be encoded

		Returns:
			str
		"""

		# If we have a datetime object
		if isinstance(obj, datetime):
			return obj.strftime('%Y-%m-%d %H:%M:%S')

		# Else if we have a decimal object
		elif isinstance(obj, Decimal):
			return '{0:f}'.format(obj)

		# Bubble back up to the parent default
		return json.JSONEncoder.default(self, obj)
